// EDynAllocator.h: interface for the EDynAllocator class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EDYNALLOCATOR_H__FD31CFC6_BB54_11D2_91EF_004005446EE6__INCLUDED_)
#define AFX_EDYNALLOCATOR_H__FD31CFC6_BB54_11D2_91EF_004005446EE6__INCLUDED_
#include "EAllocator.h"

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

template< class TElm, class EParentAllocator=EAllocatorBinT<TElm> >
class EDynAllocatorT : public EParentAllocator
{
public:
  /*! Constructs the allocator.
   * 
   * Set the array to NULL, and granularity to 16.
   */
	EDynAllocatorT();

  /*! Destructs the allocator.
   *
   * \warning Child class must do the own to resize(0) if the destruct() method
   * is overloaded because the overrided destruct() method will not be called at
   * this level.
   */
	virtual ~EDynAllocatorT();

  /*! Get a reference on a object of a given index.
   * @param nIndex Zero based index of the object.
   * @return Reference on the object.
   */
  virtual const TElm &at( SL nIndex ) const;

  /*! Get a reference on a object of a given index.
   * @param nIndex Zero based index of the object.
   * @return Reference on the object.
   */
  virtual TElm &at( SL nIndex );

  /*! Resizes the array so that a given number of elements fit in.
   * @param nNewSize Number of elements that should fit in the array.
   *                 If 0 then the array is freeed.
   */
  virtual void resize( SL nNewSize );

  /*! Returns the exact number of elements in the array.
   * @return Number of elements in the array.
   */
  virtual SL count() const;

  /*! Get the allocated buffer size
   * @return The maximum number of elements that can fit in the array without resizing.
   */
  virtual SL maxCount() const;

  /*! Set granularity.
   * @param nGranularity New granularity. Must be greater than 0.
   */
  virtual void setGranularity( SL nGranularity );

  /*! Get the granularity.
   * @return Granularity.
   */
  virtual SL granularity() const;



private:
  /**@shapeType DependencyLink*/
  /*#  EAllocatorT lnkUnnamed*/
  TElm *_pDatas;
  SL _nCount;
  SL _nMaxCount;
  SL _nGranularity;
};



EDECLARE_ALLOCATOR( EDynAllocatorT, EAllocatorBinT, EDynAllocatorBinT );
EDECLARE_ALLOCATOR( EDynAllocatorT, EAllocatorObjT, EDynAllocatorObjT );
EDECLARE_ALLOCATOR( EDynAllocatorT, EAllocatorPtrT, EDynAllocatorPtrT );



template<class TElm,class EParentAllocator>
EDynAllocatorT<TElm,EParentAllocator>::EDynAllocatorT()
{
  _pDatas = NULL;
  _nCount = 0;
  _nMaxCount = 0;
  _nGranularity = 12;
}


template<class TElm,class EParentAllocator>
EDynAllocatorT<TElm,EParentAllocator>::~EDynAllocatorT()
{
  resize( 0 );
}


template<class TElm,class EParentAllocator>
void 
EDynAllocatorT<TElm,EParentAllocator>::resize( SL nNewSize )
{
  EASSERT( nNewSize >= 0 );

  SL nCurSize = count();

  if ( nNewSize == nCurSize )   // no change 
    return;
  
  if ( nNewSize < nCurSize )
    destruct( nNewSize, nCurSize - nNewSize );

  if ( !nNewSize )      // free datas array
  {
    if ( _pDatas )
    {
      efree( _pDatas );
      _nCount = 0;
      _pDatas = ENIL;
      _nMaxCount = 0;
    }
  }
  else                  // resize array
  {
    SL nGranularity = granularity();
    SL nIdealLen = nNewSize + ( nGranularity - nNewSize % nGranularity);
    SL nCurMaxLen = _nMaxCount;

    if ( nIdealLen != nNewSize )
    {
      _pDatas = (TElm *)erealloc( _pDatas, 
                                  sizeof( TElm ) * nNewSize );
      _nMaxCount = nIdealLen;
    }
    _nCount = nNewSize;

    if ( nCurSize < nNewSize )
    {
      construct( nCurSize,
                 nNewSize - nCurSize );
    }
  }
}


template<class TElm,class EParentAllocator>
inline SL 
EDynAllocatorT<TElm,EParentAllocator>::count() const
{
  return _nCount;
}


template<class TElm,class EParentAllocator>
inline const TElm &
EDynAllocatorT<TElm,EParentAllocator>::at( SL nIndex ) const
{
  EASSERT( nIndex >= 0  );
  EASSERT( nIndex < _nCount );
  return _pDatas[ nIndex ];
}


template<class TElm,class EParentAllocator>
inline TElm &
EDynAllocatorT<TElm,EParentAllocator>::at( SL nIndex )
{
  EASSERT( nIndex >= 0  );
  EASSERT( nIndex < _nCount );
  return _pDatas[ nIndex ];
}


template<class TElm,class EParentAllocator>
inline SL 
EDynAllocatorT<TElm,EParentAllocator>::maxCount() const
{
  return _nMaxCount;
}


template<class TElm,class EParentAllocator>
inline void 
EDynAllocatorT<TElm,EParentAllocator>::setGranularity( SL nGranularity )
{
  EASSERT( nGranularity > 0 );
  _nGranularity = nGranularity;
}

template<class TElm,class EParentAllocator>
inline SL 
EDynAllocatorT<TElm,EParentAllocator>::granularity() const
{
  return _nGranularity;
}


#endif // !defined(AFX_EDYNALLOCATOR_H__FD31CFC6_BB54_11D2_91EF_004005446EE6__INCLUDED_)
