// //////////////////////////////////////////////////////////////////////////
// Header file EHolderPtr.h for class EHolderPtr
// (c)Copyright 2000, Baptiste Lepilleur.
// Created: 2000/09/19
// //////////////////////////////////////////////////////////////////////////
#ifndef EHOLDERPTR_H
#define EHOLDERPTR_H


/*! \class EHolderPtr
 * \brief This template holds a pointer an object with a variable ownership policy.
 */
template<class ObjType>
class EHolderPtrT
{
public:
  /*! Constructs a EHolderPtr object.
   */
  EHolderPtrT();

  /*! Constructs a EHolderPtr object.
   * @param Pointer on the object.
   * @param bIsOwned Indicates if the object is owned by the holder.
   */
  EHolderPtrT( ObjType *pObject,
               bool bIsOwned = true );

  /*! Copy constructor.
   * 
   * The holder points on the same object as copy and TAKE the ownership of the
   * object.
   *
   * @param copy Object to copy.
   */
  EHolderPtrT( EHolderPtrT<ObjType> &copy );

  /*! Destructor.
   * Destroys the associated object if it is owned.
   */
  virtual ~EHolderPtrT();

  /*! Copy operator.
   * 
   * The holder points on the same object as copy and TAKE the ownership of the
   * object.
   *
   * @param copy Object to copy.
   * @return Reference on this object.
   */
  EHolderPtrT<ObjType> &operator =( EHolderPtrT<ObjType> &copy );

  /*! Returns a pointer on the object.
   * @return Pointer on the object. May be \c NULL.
   */
  ObjType *operator->() const;

  /*! Returns a pointer on the object.
   * @return Pointer on the object. May be \c NULL.
   */
  ObjType *get() const;

private:
  ObjType *_pObject;
  bool _bIsOwned;
};



// Inlines methods for EHolderPtr:
// -------------------------------


template<class ObjType>
EHolderPtrT<ObjType>::EHolderPtrT() :
    _pObject( NULL ),
    _bIsOwned( false )
{
}


template<class ObjType>
EHolderPtrT<ObjType>::EHolderPtrT( ObjType *pObject,
    bool bIsOwned ) :
    _pObject( pObject ),
    _bIsOwned( pObject  &&  bIsOwned )
{
}


template<class ObjType>
EHolderPtrT<ObjType>::EHolderPtrT( EHolderPtrT<ObjType> &copy ) :
    _pObject( copy._pObject ),
    _bIsOwned( copy._bIsOwned )
{
  copy._bIsOwned = false;
}


template<class ObjType>
EHolderPtrT<ObjType>::~EHolderPtrT()
{
  if ( _bIsOwned  &&  _pObject )
    delete _pObject;
}


template<class ObjType>
EHolderPtrT<ObjType> &
EHolderPtrT<ObjType>::operator =( EHolderPtrT<ObjType> &copy )
{
  if ( this != &copy )
  {
    if ( _pObject != copy._pObject )
    {
      if ( _bIsOwned  &&  _pObject)
        delete _pObject;
  
      _pObject = copy._pObject;
    }
    _bIsOwned = copy._bIsOwned;
    copy._bIsOwned = false;
  }

  return *this;
}


template<class ObjType>
ObjType *
EHolderPtrT<ObjType>::operator->() const
{
  EASSERT( _pObject );
  return _pObject;
}


template<class ObjType>
ObjType *
EHolderPtrT<ObjType>::get() const
{
  return _pObject;
}



#endif  // EHOLDERPTR_H
