// ###########################################################################
// File: EBITARR.HPP
// (c)Copyright 03/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// ###########################################################################


#ifndef EBITARR_HPP_INCLUDED


// Includes:
// #####################Last change: 27/03/98 22h05mn#########################

// Constantes:
// #####################Last change: 27/03/98 22h05mn#########################


// Macros:
// #####################Last change: 27/03/98 22h05mn#########################


// Types:
// #####################Last change: 27/03/98 22h05mn#########################

/*!
 * \class EBitArray
 * \brief The class EBitArray provides arrays of bit.
 * The bit can then be copied, tested, or set using high level function.
 * 
 * The bit are referenced in the array by their position. The bit of position
 * 0 is the first bit of the array.
 *
 * This class has been designed to be used with EBitStream.
 *
 * \sa EBitStream.
 */

class EBitArray
{
public:
  /*! Default Constructor.
   * The size of the array is set to 0.
   */
  EBitArray();

  /*!
   * Set a bit in the array.
   * @param nIndex Index of the bit to set. 0 is the first bit of the array.
   *               Must be in valid range (0 \< \a nIndex \< count()).
   * @param bSet If \c true then the bit is set, otherwise the bit is unset.
   */
  void set( SL nIndex, bool bSet );

  /*! Set a slice of bit in the array.
   * The given slice must fit into the array. 
   * @param nIndex Index of the first bit of the slice to set.
   *               Must be in valid range (0 \< \a nIndex \< count()).
   * @param nLength Number of bits in the slice. Must be >= 0.
   * @param bSet if \c true then the bits are set, otherwise they are unset.
   */
  void setSlice( SL nIndex, SL nLength, bool bSet );

  /*! Test if a bit is set.
   * @param nIndex Index of the bit to test.
   * @return \c true if the bit is set, \c false otherwise.
   */
  bool test( SL nIndex ) const;

  /*! Test if a bit is set.
   * @param nIndex Index of the bit to test.
   * @return \c true if the bit is set, \c false otherwise.
   */
#ifdef _MSC_VER
  bool operator[]( int nIndex ) const;
#else
  bool operator[]( SL nIndex ) const;
#endif

  /*! Get the size of the array.
   * @return size of the array.
   */
  SL count() const;

  /*! Resize the array.
   * @param nSize New size of the array, in bits.
   * @param bClear If the array is extended and \a bClear is \c true the new bits
   *        are unset, otherwise they are undefined.
   */
  void resize( SL nSize, bool bSet =true );

  /*! Replace the bit array by the copy of the \a bits array.
   * @param bits Array to copy.
   */
  void copy( const EBitArray& bits );

  /*! Replace the bit array by the copy of the \a bits array.
   * @param bits Array to copy.
   * @return This array.
   */
  EBitArray &operator =( const EBitArray &bits );

  /*! Get the internal array of byte that is used to store the bit array.
   * Using that array you can access to the content of the bit array.
   * Child class can also use bitRef() and bitMask() to manipulate the
   * array.
   *
   * The offset of the byte that contains the state of the \c nIndex 'th
   * bits of the array is given by: offset = nIndex / 8.
   *
   * The mask used to manipulate this same bit is: mask = 1 << ( nIndex & 7).
   * @return The array of bytes.
   * \sa bitRef(), bitMask(), bufferSize().
   */
  const UB *datas() const;

  /*! Get the internal array of byte that is used to store the bit array.
   * Using that array you can access to the content of the bit array.
   * Child class can also use bitRef() and bitMask() to manipulate the
   * array.
   *
   * See datas() for explanation.
   *
   * @return The array of bytes.
   * \sa bitRef(), bitMask(), bufferSize().
   */
  UB *datas();

  /*! Initialize members.
   * This methods is usally called by the constructor, but you must call it
   * If the constructor has not been call !
   */
  void init();

protected:
  /*! Returns the reference on the byte that contains a bit.
   * @param nIndex Index of the bit.
   */
  UB& bitRef( SL nIndex );

  /*! Returns the reference on the byte that contains a bit.
   * @param nIndex Index of the bit.
   */
  const UB& bitRef( SL nIndex ) const;

  /*! Returns the mask to acces a bit in a byte.
   * @param nIndex Index of the bit.
   */
  UB bitMask( SL nIndex ) const;

  /*! Get buffer size.
   * @return The size of the number of bytes used to store the bit array.
   */
  SL bufferSize() const;

  /*! Array of bits, packed on 8 bits integer.
   */
  UB *_aBits;

  /*! Length of the array in bits.
   */
  SL _nLength;
};

// Inlines:
// #####################Last change: 27/03/98 22h05mn#########################

inline UB EBitArray::bitMask( SL nIndex ) const
{
  return 1 << (nIndex & 7);
}

inline UB& EBitArray::bitRef( SL nIndex )
{
  return _aBits[ nIndex >> 3 ];
}

inline const UB& EBitArray::bitRef( SL nIndex ) const
{
  return _aBits[ nIndex >> 3 ];
}

inline void EBitArray::set( SL nIndex, bool bSet )
{
  EASSERT( nIndex >= 0 );
  EASSERT( nIndex < _nLength );
  UL nBitMask = bitMask( nIndex );
  UB &nBit = bitRef( nIndex );
  nBit &= ~ nBitMask;
  if ( bSet )
    nBit |= nBitMask;
}

inline bool EBitArray::test( SL nIndex ) const
{
  EASSERT( nIndex >= 0 );
  EASSERT( nIndex < _nLength );
  return (bitRef( nIndex ) & bitMask( nIndex )) != 0;
}

#ifdef _MSC_VER
inline bool EBitArray::operator[]( int nIndex ) const
#else
inline bool EBitArray::operator[]( SL nIndex ) const
#endif
{
  return test( nIndex );
}

inline SL EBitArray::count() const
{
  return _nLength;
}

inline SL EBitArray::bufferSize() const
{
  return ( _nLength+7) /8;
}

/*! Get the array of byte.
  * @return the array of byte.
  */
inline const UB *EBitArray::datas() const
{
  return _aBits;
}

/*! Get the array of byte.
  * @return the array of byte.
  */
inline UB *EBitArray::datas()
{
  return _aBits;
}


// Prototypes:
// #####################Last change: 27/03/98 22h05mn#########################



#define EBITARR_HPP_INCLUDED
#endif
