// ###########################################################################
// File: EFASTDEV.CPP
// (c)Copyright 03/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// Version 1.0
// ###########################################################################

/* Test:
 * - [OK] getByte()
 * - [OK] putByte()
 * - [  ] Test readMode() & writeMode() switch.
 * - [  ] Test at(x) when reading.
 * - [  ] Test atEnd()
 * - [OK] Test at()
 * - [  ] Test size()
 * - [  ] Test readBlock
 * - [--] Test writeBlock
 * - [OK] Test flush()
 * - [OK] Test close()
 */

#include <stdlib.h>
#include "usuel.hpp"
#include "efastdev.hpp"

/** Constructs a fast device for an i/o device.
  * @param pDevice pDevice the fast device read/write info from/to.
  * @param nCacheSize cache size.
  */
EFastDevice::EFastDevice( EIODevice *pDevice, SL nCacheSize )
{
  assert( nCacheSize > 64 );
  _pDevice = pDevice;
  _nCachePos = 0;
  _nCacheSize = 0;
  _nCacheStep = EBOUND( nCacheSize /32, 1, 128 );
  _nCacheMaxSize = nCacheSize + _nCacheStep;
  _pCache = (UB*)emalloc( sizeof(UB) * _nCacheMaxSize );
  assert( _pCache );
  _pEndCache = _pCache;
  _pCurByteInCache = _pCache;
  _bAccessReadMode = true;
}

/** Destructor.
  * Close device if open.
  */
EFastDevice::~EFastDevice()
{
  if ( _pDevice )
  {
    if ( isOpen() )
      close();
  }

  EIODevice::~EIODevice();
}

/** Push-back a byte on the stream.
  * @param c byte to push-back.
  * @exception EIODevice on error.
  */
void EFastDevice::ungetByte( UB c )
{
  assert( _bAccessReadMode );
  if ( _pCurByteInCache == _pCache )
  {
    shiftCache( 1 );
    _pCache[ 0 ] = c;
  }
  else
  {
    _pCurByteInCache--;
    assert( *_pCurByteInCache == c );
  }
}

/** Shift the cache buffer.
  * @param nShift number of byte to shift. If nShift > 0 then shift cache
  *        toward the right ( [0] -> [nShift] ). If nShift < 0 then shift
  *        cache toward the left ( [nShift] -> [0] ).
  */
void EFastDevice::shiftCache( SL nShift )
{
  if ( !nShift )
    return;
  if ( nShift < 0 )
    memcpy( _pCache, &_pCache[ -nShift ], _nCacheSize + nShift );
  else
    memmove( &_pCache[ nShift ], _pCache, _nCacheSize -nShift );
}

/** Step cache forward by _nCacheStep byte for reading.
  * @exception EXCIODevice.
  */
void EFastDevice::stepCacheForwardRead()
{
  assert( _bAccessReadMode );
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  SL nStep = EMAX( 0, _nCacheSize -_nCacheStep );

  shiftCache( -nStep );
  _nCachePos += nStep;

  SL nAfterStep = _nCacheSize -nStep;   // position where newly read data are stored.
  _pDevice->at( _nCachePos+ nAfterStep);
  SL nRead = _pDevice->readBlock( &_pCache[ nAfterStep ],
                                  _nCacheMaxSize -nAfterStep );
  _pCurByteInCache = &_pCache[ nAfterStep  ];
  _nCacheSize = nAfterStep + nRead;
  _pEndCache = &_pCache[ _nCacheSize ];

  if ( nRead == 0 )   // EOF !
    ETHROW( EXCIODevice( EXCIODevice::endOfFile ) );
}

/** Step cache forward by _nCacheStep byte for writing.
  * @exception EXCIODevice.
  */
void EFastDevice::stepCacheForwardWrite()
{
  assert( !_bAccessReadMode );
  assert( _pDevice );
  assert( _pDevice->isOpen() );

  SL nWrite = _nCacheSize -_nCacheStep;
  assert( nWrite > 0 );

  _pDevice->at( _nCachePos );
  _pDevice->writeBlock( _pCache, nWrite );

  _nCachePos += nWrite;
  shiftCache( -nWrite );
  _nCacheSize = _nCacheMaxSize;
  _pEndCache = &_pCache[ _nCacheSize ];
  _pCurByteInCache = &_pCache[ _nCacheStep ];
}


/** Read a block of byte from the device.
  * Read \a len Byte from the device and store them in the buffer \a data.
  * @param data buffer where data are stored.
  * @param len number of bytes of data to read.
  * @return number of byte of data actually read.
  * @exception EXCIODevice is case of grave error.
  */
SL EFastDevice::readBlock( void *data, SL nLen)
{
  assert( _bAccessReadMode );
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  assert( nLen >= 0 );

  SL nTotalRead = 0;
  SL nLeft = _pEndCache - _pCurByteInCache;
  SL nFirstRead = EMIN( nLeft, nLen );
  memcpy( data, _pCurByteInCache, nFirstRead );   // read til end of cache

  nLen -= nFirstRead;
  _pCurByteInCache += nFirstRead;
  data = (void*)( ((UB*)data) + nFirstRead );
  nTotalRead += nFirstRead;

  if ( nLen )       // still left some to be read !
  {
    _nCachePos += _nCacheSize;
    _pDevice->at( _nCachePos );
    if ( nLen > _nCacheMaxSize )    // read from device to data buffer
    {
      SL nRead = _pDevice->readBlock( data, nLen );
      _nCachePos += nRead;
      nTotalRead += nRead;
      nLen = 0;
    }
                                    // read to cache
    SL nRead = _pDevice->readBlock( _pCache, _nCacheMaxSize );
    _nCacheSize = nRead;
    _pEndCache = &_pCache[ _nCacheSize ];
    _pCurByteInCache = _pCache;

    if ( nLen )                     // read from cache to data buffer
    {
      SL nRead = EMIN( nLen, _nCacheSize );
      memcpy( data, _pCache, nRead );
      _pCurByteInCache += nRead;
      nTotalRead += nRead;
    }
  }
  return nTotalRead;
}

/** Write a block of byte on the device.
  * Write \a len Byte of buffer \a data on the device.
  * @param data buffer where data are read.
  * @param len number of bytes of data to write.
  * @exception EXCIODevice is case of grave error.
  */
void EFastDevice::writeBlock( const void *data, SL nLen )
{
  assert( !_bAccessReadMode );
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  assert( nLen >= 0 );

  SL nLeft = _pEndCache - _pCurByteInCache;
  SL nFirstWrite = EMIN( nLeft, nLen );
  memcpy( _pCurByteInCache, data, nFirstWrite );   // write til end of cache

  nLen -= nFirstWrite;
  _pCurByteInCache += nFirstWrite;
  data = (void*)( ((UB*)data) + nFirstWrite );

  if ( nLen )       // still left some to be written !
  {
    _pDevice->at( _nCachePos );      // flush cache
    _pDevice->writeBlock( _pCache, _nCacheSize );
    _nCachePos += _nCacheSize;

    _pDevice->writeBlock( data, nLen );    // write others datas to device
    _nCachePos += nLen;
    _nCacheSize = _nCacheMaxSize;
    _pEndCache = &_pCache[ _nCacheSize ];
  }
}


// Return the size of the file
SL EFastDevice::size() const
{
  assert( _pDevice );
  assert( _pDevice->isOpen() );

  return EMAX( _pDevice->size(), at() );
}

/** Return the index of the file.
  * @return index of the file.
  */
SL EFastDevice::at() const
{
  return _nCachePos + (_pCurByteInCache - _pCache);
}

/** Set the index of the file.
  * If \a pos is beyond end of file, then the file is expanded.
  * The new region is initialised with 0.
  * @param pos position of the index in the file.
  */
void EFastDevice::at( SL nPos )
{
  if ( _bAccessReadMode )    // do not flush if read mode & in cache buffer.
  {
    SL nCacheOfs = nPos -_nCachePos;
    if ( nCacheOfs < _nCacheSize )
      _pCurByteInCache = &_pCache[ nCacheOfs ];
    else
      startRead( nPos );
  }
  else
  {
    flush();
    startWrite( nPos );
  }
}

/** Test if the index is at the end of the file.
  * @return true if the index is at the end of the file.
  */
bool EFastDevice::atEnd() const
{
  return at() >= size();
}

// Flush the file.
void EFastDevice::flush()
{
  assert( _pDevice->isOpen() );
  if ( !_bAccessReadMode )
  {
    _pDevice->at( _nCachePos );
    _pDevice->writeBlock( _pCache, _pCurByteInCache -_pCache );
  }
}


/** Open the file.
  * Open the file of name given on construction or with \l setName() using
  * mode \a _mode.
  * @param _mode mode used to open the file.
  * @see EIODevice::flags()
  * @exception EXCIODevice raised on error
  */
void EFastDevice::open( int _mode )
{
  assert( _pDevice );
  assert( !_pDevice->isOpen() );

  resetInfo();
  setMode( _mode );

  assert( !isReadable()  ||  !isWritable() );

  _pDevice->open( _mode );
  setStatus( EIO_IsOpen );
  if ( !isReadWrite()  &&  isWritable() )
  {
    _bAccessReadMode = false;
    startWrite( 0 );
  }
  else
  {
    _bAccessReadMode = true;
    startRead( 0 );
  }
}

// Close the file.
void EFastDevice::close()
{
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  flush();
  _pDevice->at( at() );
  _pDevice->close();
  unsetStatus( EIO_IsOpen );
  resetInfo();
}


/// Set access as exclusive read mode.
void EFastDevice::readMode()
{
  if ( _bAccessReadMode )
    return;
  _bAccessReadMode = true;
  startRead( at() );
}

/// Set access as exclusive read mode.
void EFastDevice::writeMode()
{
  if ( !_bAccessReadMode )
    return;
  _bAccessReadMode = false;
  startWrite( at() );
}

/** Prepare cache for reading.
  * @param nPos position in file.
  */
void EFastDevice::startRead( SL nPos )
{
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  _pDevice->at( nPos );
  _nCacheSize = _pDevice->readBlock( _pCache, _nCacheMaxSize );
  _pEndCache = &_pCache[ _nCacheSize ];
  _nCachePos = nPos;
  _pCurByteInCache = _pCache;
}

/** Prepare cache for writing.
  * @param nPos position in file.
  */
void EFastDevice::startWrite( SL nPos )
{
  assert( _pDevice );
  assert( _pDevice->isOpen() );
  _nCachePos = nPos;
  _nCacheSize = _nCacheMaxSize;
  _pCurByteInCache = _pCache;
  _pEndCache = &_pCache[ _nCacheSize ];
}

/** Attach a device to the fast file.
  * @param pDevice pDevice the fast device read/write info from/to.
  */
void EFastDevice::setDevice( EIODevice *pDevice )
{
  if ( _pDevice )   // there is another device attached !
  {
    assert( !_pDevice->isOpen() );
    _pDevice->close();
  }
  _pDevice = pDevice;
}

// ########################### EOF EFASTDEV.CPP ##############################
// ########################### EOF EFASTDEV.CPP ##############################
// ########################### EOF EFASTDEV.CPP ##############################

