// ###########################################################################
// File: EHASH.HPP
// (c)Copyright 03/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// ###########################################################################


#ifndef EHASH_HPP_INCLUDED

/* History:
 * 26/07/98: - Add typedef in EHash for Iterator class.
 *           - EHashIterator declaration is now included in this file.
 *           - Add isEmpty().
 * 13/09/1998: BugFix: lookUp() const; Non finite recursity ! => copy/paste non const method.
 */



// Includes:
// #####################Last change: 27/03/98 21h49mn#########################
#include "EDynArray.h"

// Constantes:
// #####################Last change: 27/03/98 21h49mn#########################


// Macros:
// #####################Last change: 27/03/98 21h49mn#########################


// Types:
// #####################Last change: 27/03/98 21h49mn#########################

/*!
 * \brief Template class that provides an hash table with complex key.
 *
 * The template definition is as follow:
 *
 * \code
 * template<class Key, class Item, class KeyArg, class ItemArg>
 * class EHash;
 * \endcode
 *
 * @param Key Type of used for key. This type is used for key storage and 
 *            for return values.
 * @param Item Type of the item associated with a key. This type is used 
 *            for key storage and for return values.
 * @param KeyArg Type for key when used as parameter. It is usually
 *               a reference on the \a Key type.
 * @param ItemArg Type for item when used as parameter. It is usually
 *               a reference on the \a Item type.
 *
 * The \a Key and \a Item type must provides a copy operator and constructor.
 *
 * To use this class you must subclass it and overide the virtual method
 * getKey() that compute a integer value based on the key and compareKey()
 * that compare two keys.
 *
 * If the value used for the key is already numeric then you should consider
 * using EHashVal.
 *
 * How does this works ?
 * 
 * An hash table stores association of (key, item). The must be unique among
 * all the associations. The key is used as an identifier of the item and
 * provides a way to store a large amount data and retreive them using
 * the key only.
 *
 * Association are added to the table using add(). They can be later retreived
 * using operator[] or lookUp() if you need to test the existence of a key.
 * 
 * Association are removed from the table using remove() or removeAll().
 * You can overide freeAsso() if you need to do some special clean up when 
 * association are destroyed. Note that destructor are called for the key
 * and the item. freeAsso() is needed in some case, for example when the key
 * is an allocated string of char *, you need to free the memory in this method.
 *
 * Efficiency notes:
 *
 * You should try to keep the table always 20%-30% greater than the number
 * of associations it contains.
 *
 *
 * Example of use:
 *
 * In this example, we need a hash table to compute frequency of words in a
 * text.
 *
 * Words are stored using EString and are used as key. The frequency is
 * the item and is an integer.
 *
 * \code
 *
 *   class HashWordFrq : public EHash<char *,UL,char *,UL>
 *   {
 *   public:
 *     // This method must return a numeric value based on the key.
 *     UL getKey( EString &key ) const
 *     {
 *       return key.makeHashKey();    // EString provides a method to compute
 *                                    // a numeric value based on the string.
 *     }
 *
 *     // This method must test if two keys are equal.
 *     bool compareKey( EString &key1, EString &key2 ) const
 *     {
 *       return key1 == key2;
 *     }
 *   };
 *
 *   // A method that computes frequency:
 *   // reader is an object that gives words from the text on by one.
 *   // hash is an hash table used to store words frequency !
 *   void computeFrq( TextReader &reader, 
 *                    HashWordFrq &hash )
 *   {
 *     while ( !reader.atEnd() )    // While there is words to read...
 *     {
 *       EString strWord = reader.getWord();    // Read the word.
 *       if ( !hash.lookUp( strWord ) )         // Is the word in the table ?
 *       {                                         // No !
 *         hash.add( strWord, 1 );                   // => add an association.
 *       }
 *       else
 *         hash[ strWord ]++;                      // Yes: increase its frequency !
 *     }
 *   }
 * \endcode
 *
 *  Notes: The inner loop of the example is not optimized as there is 2 lookups:
 *  The first when testing if the key exist and the second one when increasing
 *  the frequency. Only the first is needed but I needed this for the purpose
 *  of the example.
 *  \sa EHashIterator, EHashVal, EHashValIterator.
 */


template<class Key, class Item, class KeyArg, class ItemArg>
class EHash
{
#ifdef _SUPPORT_TEMPLATE_FRIEND
  friend class EHashIterator<Key,Item,KeyArg,ItemArg>;
#endif

public:
  /*! Constructor.
   * @param nSize initial size of the table.
   */
  EHash( SL nSize = 3 );

  /*! Destructor
   */
  virtual ~EHash();

  /*! Lookup a item using its key.
   * @param key key associated to the item.
   * @return reference on item.
   */
  Item& operator[]( KeyArg key );

  /* Lookup a item using its key.
   * @param key key associated to the item.
   * @return reference on item.
   */
//  const Item& operator[]( const KeyArg key ) const;

  /*! Lookup a item using its key.
   * @param key key associated to the item.
   * @return pointeur on item if found. NULL if no item associated to key.
   */
  Item *lookUp( KeyArg key );

  /*! Lookup a item using its key.
   * @param key key associated to the item.
   * @return pointeur on item if found. NULL if no item associated to key.
   */
  const Item *lookUp( KeyArg key ) const;

  /*! Add an item to the association table.
   * @param key associated to the item.
   * @param item item to add.
   */
  void add( KeyArg key, ItemArg item );

  /*! Remove an item from the association table.
   * @param key key associated to the item to remove.
   */
  void remove( KeyArg key );

  /*! Remove all the item from the association table.
   */
  void removeAll();

  /*! Build a numeric value from the key.
   * @param key key from which the value is build.
   * @return numeric value.
   */
  virtual UL getKey( KeyArg key ) const =0;

  /*! Test if two key are equal.
   * @param key1 key to compare to key2.
   * @param key2 key to compare to key1.
   * @return \c true if the key are equal, \c false otherwise.
   */
  virtual bool compareKey( KeyArg key1, KeyArg key2 ) const =0;

  /*! Free memory use by a key and an item (in case they use indirect data).
   * For example if key is a "char *" then you need to do "free( key )".
   * @param key  key to free
   * @param item item item to free.
   */
  virtual void freeAsso( Key &key, Item &item ) {}

  /*! Resize the table.
   * @param nSize new table size.
   */
  void resize( SL nSize );

  /*! Get the size of the table.
   * @return size of the table.
   */
  SL size() const;

  /*! Get the number of association in the table.
   * @return number of association in the table.
   */
  SL count() const;

  /*! Test if the table is empty.
   * @return true if there is no association in the table, false otherwise.
   */
  bool isEmpty() const;

#ifdef _SUPPORT_TEMPLATE_FRIEND
private:
#endif
  struct ELCAsso
  {
    ELCAsso *_pNext;
    Key _key;
    Item _item;
  };

  EArray<ELCAsso *> _aAsso;

  SL _nNbAsso;
};


// Template Implementation:
// #####################Last change: 27/03/98 21h49mn#########################

/* Constructs an hash table object.
 * @param nSize size of the hash table. To avoid collision, the rules of the
 *        thumb is that nSize should be 1.3 times the number of associations.
 *        Giving large size does not use lot of memory (only 1 (void *) ptr by
 *        unit).
 */
template<class Key, class Item, class KeyArg, class ItemArg>
EHash<Key,Item,KeyArg,ItemArg>::EHash( SL nSize )
{
  EASSERT( nSize > 0 );
  _nNbAsso = 0;
  resize( nSize );
}
// Destructor
template<class Key, class Item, class KeyArg, class ItemArg>
EHash<Key,Item,KeyArg,ItemArg>::~EHash()
{
  removeAll();
}

/* Lookup a item using its key.
  * @param key key associated to the item.
  * @return pointeur on item if found. NULL if no item associated to key.
  * */
template<class Key, class Item, class KeyArg, class ItemArg>
Item *EHash<Key,Item,KeyArg,ItemArg>::lookUp( KeyArg key )
{
  SL nIndex = getKey( key ) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( compareKey( pAsso->_key, key ) )
      return &pAsso->_item;
    pAsso = pAsso->_pNext;
  }
  return NULL;
}


/* Lookup a item using its key.
  * @param key key associated to the item.
  * @return pointeur on item if found. NULL if no item associated to key.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
const Item *EHash<Key,Item,KeyArg,ItemArg>::lookUp( KeyArg key ) const
{
  SL nIndex = getKey( key ) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( compareKey( pAsso->_key, key ) )
      return &pAsso->_item;
    pAsso = pAsso->_pNext;
  }
  return NULL;
}


/* Lookup a item using its key.
  * @param key key associated to the item.
  * @return reference on item.
  * */
template<class Key, class Item, class KeyArg, class ItemArg>
Item& EHash<Key,Item,KeyArg,ItemArg>::operator[]( KeyArg key )
{
  Item *pItem = lookUp( key );
  EASSERT( pItem );
  return *pItem;
}

/* Lookup a item using its key.
  * @param key key associated to the item.
  * @return reference on item.
  * */
/*
template<class Key, class Item, class KeyArg, class ItemArg>
const Item& EHash<Key,Item,KeyArg,ItemArg>::operator[]( const Key key ) const
{
  const Item *pItem = lookUp( key );
  EASSERT( pItem );
  return *pItem;
}
*/

/* Add an item to the association table.
  * @param key associated to the item.
  * @param item item to add.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHash<Key,Item,KeyArg,ItemArg>::add( KeyArg key,
                                          ItemArg item )
{
  EASSERT( _aAsso.count() > 0 );

  SL nIndex = getKey( key ) % _aAsso.count();

  ELCAsso *pAsso = new ELCAsso();
  pAsso->_key = key;
  pAsso->_item = item;
  pAsso->_pNext = _aAsso[ nIndex ];
  _aAsso[ nIndex ] = pAsso;
  _nNbAsso++;
}


/* Remove an item from the association table.
  * @param key key associated to the item to remove.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHash<Key,Item,KeyArg,ItemArg>::remove( KeyArg key )
{
  EASSERT( _aAsso.count() > 0 );

  SL nIndex = getKey( key ) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  ELCAsso **pPrevAsso = &_aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( compareKey( pAsso->_key, key ) )  // found !
    {
      *pPrevAsso = pAsso->_pNext;
      freeAsso( pAsso->_key, pAsso->_item );
      delete pAsso;
      _nNbAsso--;
      return;
    }

    pPrevAsso = &pAsso->_pNext;
    pAsso = pAsso->_pNext;
  }
  EASSERT( pAsso );
}


/* Resize the table.
  * @param nSize new table size.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHash<Key,Item,KeyArg,ItemArg>::resize( SL nSize )
{
  EDynArray<ELCAsso *> aAsso2;
  SL nCpt;
  ELCAsso *pNull = ENIL;

  aAsso2.resize( nSize );
  aAsso2.fill( &pNull );

  for ( nCpt = 0; nCpt < size(); ++nCpt )
  {
    ELCAsso *pAsso = _aAsso[ nCpt ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      SL nIndex = getKey( pAsso->_key ) % nSize;    // compute new index

      pAsso->_pNext = aAsso2[ nIndex ];              // add to head of lc
      aAsso2[ nIndex ] = pAsso;

      pAsso = pNextAsso;
    }
  }

  _aAsso = aAsso2;
}


template<class Key, class Item, class KeyArg, class ItemArg>
inline SL EHash<Key,Item,KeyArg,ItemArg>::size() const
{
  return _aAsso.count();
}


template<class Key, class Item, class KeyArg, class ItemArg>
inline SL EHash<Key,Item,KeyArg,ItemArg>::count() const
{
  return _nNbAsso;
}

template<class Key, class Item, class KeyArg, class ItemArg>
void EHash<Key,Item,KeyArg,ItemArg>::removeAll()
{
  SL nIndex;

  for ( nIndex = size() -1; nIndex >= 0; --nIndex )
  {
    ELCAsso *pAsso = _aAsso[ nIndex ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      freeAsso( pAsso->_key, pAsso->_item );
      delete pAsso;
      pAsso = pNextAsso;
    }
    _aAsso[ nIndex ] = ENIL;
  }
  _nNbAsso = 0;
}


/* Test if the table is empty.
 * @return true if there is no association in the table, false otherwise.
 */
template<class Key, class Item, class KeyArg, class ItemArg>
bool 
EHash<Key,Item,KeyArg,ItemArg>::isEmpty() const
{
  return count() == 0;
}


// Includes:
// #####################Last change: 27/03/98 21h49mn#########################

#include "EHashIte.hpp"


#define EHASH_HPP_INCLUDED
#endif

