// //////////////////////////////////////////////////////////////////////////
// Header file Rfta/Model/CollectionAsserter.h for class CollectionAsserter
// (c)Copyright 2000, Baptiste Lepilleur.
// Created: 2001/10/12
// //////////////////////////////////////////////////////////////////////////
#ifndef GAIACRTN_COLLECTIONASSERTER_H
#define GAIACRTN_COLLECTIONASSERTER_H


#include <algorithm>
#include <deque>
#include <cppunit/Asserter.h>
#include "Assert.h"


namespace Assert
{

template<typename ColType>
std::string collectionToString( const ColType &collection )
{
  std::string message( "{ " );
  ColType::const_iterator it = collection.begin();
  while ( it != collection.end() )
  {
    message += toString( *it );
    ++it;
    if ( it != collection.end() )
      message += ",\n";
  }
  message += " }";
  return message;
}




template<typename ExpectedCol,
         typename ActualCol>
struct CollectionContainsSameElementsChecker
{
  typedef ActualCol::value_type ItemType;
  typedef std::deque<ActualCol::value_type> Items;

  CollectionContainsSameElementsChecker( const ExpectedCol &expectedCol,
                                         const ActualCol &actualCol ) :
      _expectedCol( expectedCol ),
      _actualCol( actualCol )
  {
  }


  void check( const CppUnit::SourceLine &sourceLine )
  {
    if ( !hasMissingOrExtraneousItems() )
      return;

    reportFailure( sourceLine );
  }

  
  bool hasMissingOrExtraneousItems()
  {
    std::copy( _actualCol.begin(), _actualCol.end(), 
               std::back_inserter( _extraneousActualItems ) );

    ExpectedCol::const_iterator itExpected( _expectedCol.begin() );
    while ( itExpected != _expectedCol.end() )
    {
      Items::iterator itFound = std::find( _extraneousActualItems.begin(),
                                           _extraneousActualItems.end(),
                                           *itExpected );
      if ( itFound != _extraneousActualItems.end() )
        _extraneousActualItems.erase( itFound );
      else
        _missingExpectedItems.push_back( *itExpected );

      ++itExpected;
    }

    return !_missingExpectedItems.empty()  ||
           !_extraneousActualItems.empty();
  }


  void reportFailure( const CppUnit::SourceLine &sourceLine )
  {
    std::string expected( collectionToString( _expectedCol ) );
    std::string actual( collectionToString( _actualCol ) );
    std::string message( "Missing expected items:\n" );
    message += collectionToString( _missingExpectedItems );
    message += "\nExtraneous actual items:\n";
    message += collectionToString( _extraneousActualItems );
    message += '\n';

    CppUnit::Asserter::failNotEqual( expected, 
                                     actual, 
                                     sourceLine, 
                                     message );
  }


  const ExpectedCol &_expectedCol;
  const ActualCol &_actualCol;
  Items _extraneousActualItems;
  Items _missingExpectedItems;
};


template<typename ExpectedCol,
         typename ActualCol>
void 
checkCollectionContainsSameElements( const ExpectedCol &expectedCol,
                                     const ActualCol &actualCol,
                                     const CppUnit::SourceLine &sourceLine )
{
  CollectionContainsSameElementsChecker<ExpectedCol,ActualCol> checker( 
                                                                expectedCol,
                                                                actualCol );
  checker.check( sourceLine );
}



}  //  namespace Assert



#define GAIACRTN_ASSERT_SAME_COLLECTION( expected, actual )            \
    Assert::checkCollectionContainsSameElements( expected,             \
                                                 actual,               \
                                                 CPPUNIT_SOURCELINE() )


#endif  // GAIACRTN_COLLECTIONASSERTER_H
