// //////////////////////////////////////////////////////////////////////////
// Header file TtrWriterWrappedTxt.h for class TtrWriterWrappedTxt
// (c)Copyright 2000, Baptiste Lepilleur.
// Created: 2000/09/27
// //////////////////////////////////////////////////////////////////////////
#ifndef TTRWRITERWRAPPEDTXT_H
#define TTRWRITERWRAPPEDTXT_H

#include "TtrFmtEltVisitor.h"
class TtrWriterSetting;


/*! \class TtrWriterWrappedTxt
 * \brief This class represents a writer that writes a wrapped formatted text in a text file.
 */
class TtrWriterWrappedTxt : public TtrFmtEltVisitor
{
  friend class TtrUtWriterWrappedTxt;
public:
  /*! Constructs a TtrWriterWrappedTxt object.
   * @param rstreamOut Output text stream.
   * @param setting Writer setting.
   */
  TtrWriterWrappedTxt( ETextStream &rstreamOut,
                       const TtrWriterSetting &setting );

  /*! Destructor.
   */
  virtual ~TtrWriterWrappedTxt();

  /*! Visits a blank block.
   * @param blank Blank block.
   */
  virtual void visitBlank( TtrFmtBlank &blank );

  /*! Visits a paragraph.
   * @param para Paragraph.
   */
  virtual void visitPara( TtrFmtPara &para );

  /*! Should be called when the writer is started.
   */
  void start() {}

  /*! Should be called when the writer is done.
   */
  void end() {}

  /*! Returns the text stream used by the writer.
   * @return Text stream used by the writer.
   */
  ETextStream &stream() const;

private:
  struct BreakPos
  {
    BreakPos( SL nPos, SL nLength ) : _nPos( nPos ), _nLength( nLength ) {}
    SL _nPos;
    SL _nLength;
  };

  typedef EBinDynArrayT<BreakPos> ArrayBreakPos;

  typedef EObjDynArrayT<EString> Strings;

private:
  /*! Prevents the use of the copy constructor.
   */
  TtrWriterWrappedTxt( const TtrWriterWrappedTxt &copy );

  /*! Prevents the use of the copy operator.
   */
  TtrWriterWrappedTxt &operator =( const TtrWriterWrappedTxt &copy );

  void setup( const TtrWriterSetting &setting );

  void writeCenteredPara( const EString &strPrefix,
                          const EString &strParagraph );
  void writeLeftAlignedPara( const EString &strPrefix,
                             const EString &strParagraph );

  /*! Finds all the word-break position in the specified text.
   *
   * A break position is added for each block of space found in the text.
   * The position of the first character of the block and the block length
   * is specified and added to \a vBreakPos.
   *
   * A last break position of length 0 is added for the end of the string
   * (even if the string is empty).
   *
   * @param strText Text in which word break position are searched.
   * @param vBreakPos [Out] Array in which word break position are added.
   */
  void findBreakPos( const EString &strText,
                     ArrayBreakPos &vBreakPos ) const;

  bool isWordChar( char c ) const;

  SL computeRightMarginPos( const EString &strLine,
                            SL nSliceFrom,
                            SL nSliceTo,
                            SL nLeftMarginPos ) const;

  /*! Breaks a paragraph into lines.
   * @param strParagraph [In] Paragraph to wrap.
   * @param lines [Out] Lines of the paragraph are added to this array.
   * @param nLineLength [In] Right margin used to wrap the paragraph.
   * @param nLeftMarginPos [In] Position of the left column. 0 based.
   */
  void breakParagraph( const EString &strParagraph, 
                       Strings &lines, 
                       int nLineLength,
                       int nLeftMarginPos ) const;

  /// Replaces each tab character with a space.
  EString replaceTabWithSpace( const EString &strParagraph ) const;
private:
  /*! Should we keep the original indentation on indented line ?
   */
  bool _bKeepOriginalIndent;

  /*! String for an indentation level "step".
   */
  EString _strIndentStep;

  /*! Reference on the output text stream.
   */
  ETextStream &_rstreamOut;

  /*! Columns at which the text is wrapped.
   */
  SL _nWrappingCol;
};



// Inlines methods for TtrWriterWrappedTxt:
// ----------------------------------------


inline ETextStream &
TtrWriterWrappedTxt::stream() const
{
  return _rstreamOut;
}



#endif  // TTRWRITERWRAPPEDTXT_H
